/*----------------------------------------------------------------------*/
/* base board for Z80-CPU                                  2012.10.4    */
/*                                for mega64A  Fuse L:H:X = EF:C1:FF    */
/*                                                                      */
/*                                                      (C)neko Java    */
/*----------------------------------------------------------------------*/


#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include "uart.h"
#include "diskio.h"
#include "aki80cpm.h"
#include "iicrom.h"
#include "pcwrite.h"


/*---------------------------------------------------------*/
/* global variables                                        */
/*---------------------------------------------------------*/

BYTE Block_buff[BLOCK_SIZE];	/* SDC block buffer */

volatile BYTE DMA_mode;		/* READ or WRITE or etc. */
volatile BYTE DMA_result;
volatile BYTE Adrl;
volatile BYTE Adrh;
volatile WORD Adr;
volatile BYTE Sect_no;
volatile BYTE Track_no_l;
volatile BYTE Track_no_h;
volatile WORD Track_no;
volatile DWORD Block_no;
volatile DWORD Pre_block_no;
volatile DWORD Sdc_clst_no;
volatile BYTE Buff_offset;
volatile BYTE Size_kb;	// IIC-EEPROM write size[kB]

volatile BYTE F_unwritten;
volatile DWORD Pre_sdc_clst_no;
volatile WORD Write_back_timer;

volatile DSTATUS St;

volatile WORD Timer;		/* 100Hz increment timer */


/*---------------------------------------------------------*/
/* 100Hz timer interrupt generated by OC2                  */
/*---------------------------------------------------------*/

ISR(TIMER2_COMP_vect)
{
	Timer++;		/* Performance counter for this module */
	disk_timerproc();	/* Drive timer procedure of low level disk I/O module */

}


/*---------------------------------------------------------*/

void IoInit ()
{
	PORTA = 0xFF;		 //ProtA :address bus low byte
	DDRA  = 0;
	PORTF = 0xFF;		 //ProtF :address bus high byte
	DDRF  = 0;

	PORTE = 0b11110100;	 //PortE :Z80 control pins
	DDRE  = 0b00011100;

	PORTB = 0b11110001;	 //PortB :Z80 control pins
	DDRB  = 0b01000001;

	PORTC = 0xFF;		 //ProtC :data bus
	DDRC  = 0;

	uart_init();		// Initialize UART driver

	iicROM_init();		// external IIC-EEPROM driver

	cli();

	OCR2 = 156-1;		// Timer2: 100Hz interval (OC2), 156-1 at 16MHz
	TCCR2 = 0b00001101;	// CTC compA Mode. 1/1024
	TIMSK = 0b10000000;	// Enable TC2.oc, interrupt

	TCCR1A = 0b00010000;	// OC1B pin toggle output. (for Z80 CPU clock)
	TCCR1B = 0b00001001;	// CTCmode. clkI/O (1/1)
	OCR1AH = 0;
	OCR1AL = 1; // 0: 8MHz, 1: 4MHz, 2: 2.6MHz, 3: 2MHz, 7: 1MHz. f=16/(2*(1+OCR1A)) [MHz]

}


/*---------------------------------------------------------*/

void nop_wait(WORD count)
{
	WORD i;
	for(i=0; i<count; i++){
		asm volatile("nop");
	}
}


void adr_set(WORD adr)
{
	BYTE high, low;
	high = (BYTE)(adr >> 8);
	low = (BYTE)(adr & 0x00FF);
	PORTF = high;
	PORTA = low;
	return;
}


void write_byte(WORD adr, BYTE dat)
{
	adr_set(adr);
	PORTC = dat;
	asm volatile("nop"); asm volatile("nop"); asm volatile("nop");
	WR_L(); asm volatile("nop"); asm volatile("nop"); asm volatile("nop");
	WR_H(); asm volatile("nop"); asm volatile("nop"); asm volatile("nop");
	return;
}


BYTE read_byte(WORD adr)
{
	BYTE dat;
	adr_set(adr);
	RD_L(); asm volatile("nop"); asm volatile("nop"); asm volatile("nop");
	dat = PINC;
	RD_H(); asm volatile("nop"); asm volatile("nop"); asm volatile("nop");
	return dat;
}


void ram_write(WORD adr, WORD size, BYTE *buff, BYTE offset)	// AVR buffer -> Z80 RAM
{
	WORD i;

	D_BUS_OUT();
	for(i=0; i<size; i++){
		write_byte(adr+i, buff[i+offset*SECT_SIZE]);
	}
	D_BUS_IN();

}


void ram_read(WORD adr, WORD size, BYTE *buff, BYTE offset)	// AVR buffer <- Z80 RAM
{
	WORD i;

	for(i=0; i<size; i++){
		buff[i+offset*SECT_SIZE] = read_byte(adr+i);
	}
}


void var_calc() // recalculation global variables 
{
	DWORD cpm_clst_no = 0;

	Adr = ((WORD)Adrh << 8) + (WORD)Adrl;

	//Track_no_h = 0;
	Track_no = ((WORD)Track_no_h << 8) + (WORD)Track_no_l;

	Pre_block_no = Block_no;
	Pre_sdc_clst_no = Sdc_clst_no;

	cpm_clst_no = (DWORD)(Sect_no + Track_no * SECT_CNT);	//CPM128oCgPʂŌNX^No.
	Block_no = cpm_clst_no / CPM_CLST_CNT_PER_BLOCK;	//SDC̏Pʂł̃ubNNo.
	Buff_offset = cpm_clst_no % CPM_CLST_CNT_PER_BLOCK;	// 0`7(1ubN1024oCgA1ZN^128oCg̏ꍇ)

	Sdc_clst_no = Block_no * SDC_CLST_CNT_PER_BLOCK;	//SDC̃NX^No.(1NX^512oCg)
}


void sdc_write(){	// Z80 RAM -> AVR buffer -> SDC
	DRESULT rs;
	var_calc();
	if(Block_no != Pre_block_no){	//ubNړꍇ
		if(F_unwritten){	//݂̏ꍇ܂ÕubNSDC֏ł
			rs = disk_write (0, Block_buff, Pre_sdc_clst_no, SDC_CLST_CNT_PER_BLOCK);
			if(rs != RES_OK){
				DMA_result = DMA_NG;
				St = STA_NOINIT;
				return;
			}
		}
		rs = disk_read (0, Block_buff, Sdc_clst_no, SDC_CLST_CNT_PER_BLOCK);	//SDCobt@֓Ǎ
		if(rs != RES_OK){
			DMA_result = DMA_NG;
			St = STA_NOINIT;
			return;
		}
	}
	ram_read(Adr, SECT_SIZE, Block_buff, Buff_offset);
	DMA_result = DMA_OK;
	F_unwritten = 1;	//݃tOĂ
	Write_back_timer = 0;	//^C}Zbg
}


void sdc_read(){	// Z80 RAM <- AVR buffer <- SDC
	DRESULT rs;
	var_calc();
	if(Block_no != Pre_block_no){	//ubNړꍇ
		if(F_unwritten){	//݂̏ꍇ܂ÕubNSDC֏ł
			rs = disk_write (0, Block_buff, Pre_sdc_clst_no, SDC_CLST_CNT_PER_BLOCK);
			if(rs != RES_OK){
				DMA_result = DMA_NG;
				St = STA_NOINIT;
				return;
			}
			F_unwritten = 0;	//݃tO낷
		}
		rs = disk_read (0, Block_buff, Sdc_clst_no, SDC_CLST_CNT_PER_BLOCK);	//SDCobt@֓Ǎ
		if(rs != RES_OK){
			DMA_result = DMA_NG;
			St = STA_NOINIT;
			return;
		}
	}
	ram_write(Adr, SECT_SIZE, Block_buff, Buff_offset);
	DMA_result = DMA_OK;
}


void rom_load(){	// AVR EEPROM -> Z80 RAM (for Z80 MONITOR program)

	BYTE count, i;
	WORD size;

	if(EEPROM_SIZE >= BLOCK_SIZE){
		count = EEPROM_SIZE/BLOCK_SIZE;
		size = BLOCK_SIZE;
	}else{
		count = 1;
		size = EEPROM_SIZE;
	}
	for(i=0; i<count; i++){
		eeprom_busy_wait();
		eeprom_read_block(Block_buff, (void *)(i*size), size); 
		eeprom_busy_wait();
		ram_write(i*size, size, Block_buff, 0);
	}
}


BYTE ext_rom_load()	// external IIC-EEPROM -> Z80 RAM(0000-7FFF)
{
	BYTE r, data;
	WORD adr;
	
	D_BUS_OUT();
	if((r=iicROM_read_start(0, 0))!=0) goto EXIT;
	for(adr=0x0000; adr<0x7FFF; adr++){
		if((r=iicROM_read(&data))!=0) goto EXIT;
		write_byte(adr, data);
	}
	if((r=iicROM_read_end(&data))!=0) goto EXIT;
	write_byte(0x7FFF, data);
	D_BUS_IN();
	return 0;
	EXIT:
	D_BUS_IN();
	return r;
}


BYTE ext_rom_save(BYTE size)	// Z80 RAM(2000-) -> external IIC-EEPROM. size[kB](1..32)
{
	BYTE buff[64], r;
	WORD adr, end_adr;

	if(size>32) size=32;
	end_adr=size*1024;
	for(adr=0; adr<end_adr; adr+=64){
		ram_read(adr+0x2000, 64, buff, 0);
		//BYTE i; for(i=0; i<64;i++) buff[i]=i; //for test
		//BYTE i; for(i=0; i<64;i++) buff[i]=0xAA; //for test
		if((r=iicROM_page_write(adr, buff))!=0) return DMA_NG;
	}
	return DMA_OK;
}


void reset_proc()	// SDC first sector -> Z80 RAM and reset
{
	DRESULT rs;

	AD_BUS_OUT();
	MREQ_L();
	WR_H(); RD_H();
	RW_OUT(); nop_wait(1);

	Block_no = 0;
	Sdc_clst_no = 0;
	Track_no_h= 0;
	F_unwritten = 0;

	sei();

	St = disk_initialize (0);

	if(St & STA_NOINIT){	// SDCǍ݂ɎsꍇEEPROMvO[h
		if(ext_rom_load()) rom_load();	// VAEEPROMNGȂAAVREEPROM.
		
		Block_no = 0xFFFF; //ROM̃vOSDCǂޏꍇɔēKɂ炵Ă
		Sdc_clst_no = Block_no * SDC_CLST_CNT_PER_BLOCK;
		F_unwritten = 0;
	}else{
		rs = disk_read (0, Block_buff, Block_no, SDC_CLST_CNT_PER_BLOCK);	// get boot loader from SDC
		if(rs==RES_OK){
			ram_write(0x2000, SECT_SIZE, Block_buff, 0); // set boot loader to Z80 RAM
			Block_buff[0]=0xC3;	// jump vector set.
			Block_buff[1]=0x00;
			Block_buff[2]=0x20;
			ram_write(0, 3, Block_buff, 0);
		}else{
			if(ext_rom_load()) rom_load();
		}

	}

	cli();

/*--for test start----------------------------------
	BYTE r;
	if((r=ext_rom_save(32))!=0) uart_put('#');
	if((r=ext_rom_load())!=0) uart_put('*');
--for test end------------------------------------*/

	RW_IN();
	MREQ_Z();
	AD_BUS_IN(); asm volatile("nop");

}


void write_back(){	//CgobN
	DRESULT rs;

	rs = disk_write (0, Block_buff, Sdc_clst_no, SDC_CLST_CNT_PER_BLOCK);
	if(rs != RES_OK){
		DMA_result = DMA_NG;
		St = STA_NOINIT;
		return;
	}
	DMA_result = DMA_OK;
	F_unwritten = 0;	//݃tO낷
	Write_back_timer = 0;	//^C}Zbg
}


/*---------------------------------------------------------*/
/* when received IOREQ or BUSAK signal from Z80            */
/*---------------------------------------------------------*/

/*-------------------------
ISR(PCINT3_vect)
{
	
}
--------------------------*/


void busak_proc()
{
	AD_BUS_OUT();
	MREQ_L();
	WR_H(); RD_H();
	RW_OUT(); asm volatile("nop");

	if(DMA_mode == DMA_WRITE_BACK){
		write_back();
	}
	if(DMA_mode == DMA_WRITE){
		sdc_write();
	}
	if(DMA_mode == DMA_READ){
		sdc_read();
	}
	if(DMA_mode == DMA_WRITE_ROM){
		cli();
		DMA_result = ext_rom_save(Size_kb);
	}

	RW_IN();
	MREQ_Z();
	AD_BUS_IN(); asm volatile("nop");
}


void ioreq_proc()
{
	BYTE port_adr;

	port_adr = PINA;

	if(RD()==0 && WR()!=0){		// AVR port -> Z80
		D_BUS_OUT();
		switch (port_adr) {
		case CON_STS:
			if(uart_test()){
				PORTC = 0xFF;	// data exist
			}else{
				PORTC = 0;	// no data
			}
			break;
		case CON_IN:
			PORTC = uart_get();
			break;
		case DMA_RS:
			PORTC = DMA_result;	// set DMA result
		}
	}
	if(WR()==0 && RD()!=0){		// AVR port <- Z80
		switch (port_adr) {
		case CON_OUT:
			uart_put(PINC);
			break;
		case TRACK_SEL_L:
			Track_no_l = PINC;
			break;
		case TRACK_SEL_H:
			Track_no_h = PINC;
			break;
		case SECTOR_SEL:
			Sect_no = PINC;
			break;
		case ADR_L:
			Adrl = PINC;
			break;
		case ADR_H:
			Adrh = PINC;
			break;
		case EXEC_DMA:
			DMA_mode = PINC;	// get read or write command
			if(St & STA_NOINIT){
				sei();
				St = disk_initialize (0);
				cli();
				if(St & STA_NOINIT){
					DMA_result = DMA_NG;
					DMA_mode = DMA_NONE;
					break;
				}
			}
			BUSRQ_L();		// DMA request
			break;
		case ROM_SAVE:
			Size_kb = PINC;	// get write size[kB]
			DMA_mode = DMA_WRITE_ROM;
			BUSRQ_L();		// DMA request
		}
	}
}


/*-----------------------------------------------------------------------*/
/* Main                                                                  */
/*-----------------------------------------------------------------------*/

int main (void)
{

	IoInit();
	WAIT_L();

	nop_wait(10);
	
	while(!PCWR()) pc_write_sub();	// PC write (PC <--> I2C-EEPROM)
	
	reset_proc(); 	// store boot-loader to Z80 RAM.
	RESET_H();		// Z80 start!

	for(;;){
		while(BUSAK()!=0 && IOREQ()!=0);	// watching H to L edges
		if(IOREQ()==0){
			while(RD()!=0 && WR()!=0);
			ioreq_proc();
			WAIT_H();						// WAITJBOWbNɂ蒼M1ōēxWAITB
			//while(RD()==0);
			D_BUS_IN();
			//while(IOREQ()==0);
			nop_wait(1);					// v
			WAIT_L();						// M1WAITJ
		}
		if(BUSAK()==0){
			sei();
			busak_proc();
			cli();
			DMA_mode = DMA_NONE;
			BUSRQ_H();	// DMA end.
			while(BUSAK()==0);
		}
		if(DMA_mode==DMA_NONE){
			if(F_unwritten){
				Write_back_timer++;
				if(Write_back_timer > 100){
					DMA_mode = DMA_WRITE_BACK;
					BUSRQ_L();
				}
			}
		}
	}
}
